use strict;
use XML::Parser;
use Digest::SHA1 qw(sha1_hex);

our (%FXim, %FXbot, %FXirc, %FXserv, %FXboard, %FXcfg, %FXflags, %FXsock, %FXmsg, $ts);

my %jabber_error = (
  400 => "Bad Request",
  401 => "Unauthorized",
  402 => "Payment Required",
  403 => "Forbidden",
  404 => "Not Found",
  405 => "Not Allowed",
  406 => "Not Acceptable",
  407 => "Registration Required",
  408 => "Request Timeout",
  409 => "Conflict",
  500 => "Internal Server Error",
  501 => "Not Implemented",
  502 => "Remove Server Error",
  503 => "Service Unavailable",
  504 => "Remove Server Timeout",
  510 => "Disconnected"
);

sub jabber_connect {
  my ($server, $port, $reconnect) = (shift, shift, shift);

  $FXim{'jabber'}{'attempt'} = $ts;
  $FXim{'jabber'}{'ping'} = $ts;
  $FXim{'jabber'}{'board'} = ();
  $FXsock{'jabber'} = "";

  if (!$reconnect) {
    $FXim{'jabber'}{'id'} = "";
    $FXim{'jabber'}{'session'} = 1;
  }

  my $socket = &open_socket("tcp", $server, $port, 3);

  if (!$socket) {
    &putinfo("y", "Unable to connect to Jabber server %s:%d", $server, $port);
    return(0);
  }

  &add_socket($socket, "JABBER");

  &putinfo("y", "Connected to Jabber server %s:%d", $server, $port);

  &jabber_output($FXsock{'jabber'}, "<?xml version='1.0'?><stream:stream xmlns='jabber:client' xmlns:stream='http://etherx.jabber.org/streams' to='%s'>", $server);
  return(0);
}

sub jabber_output {
  my ($socket, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  print $socket $buffer;
  return(0);
}

sub jabber_id {
  my $method = shift;
  my $username = (split(/\@/, $FXbot{'PageServ'}{'jabber_id'}, 0))[0];

  if ($method eq "digest") {
    my $digest = sha1_hex($FXim{'jabber'}{'id'} . $FXbot{'PageServ'}{'jabber_password'});
    &jabber_output($FXsock{'jabber'}, "<iq type='set' id='doConnect'><query xmlns='jabber:iq:auth'><username>%s</username><digest>%s</digest><resource>%s</resource></query></iq>", $username, $digest, "FXserv");
    return(0);
  }

  return(0);
}

sub jabber_parse {
  my $buffer = shift;

  $FXim{'jabber'}{'ping'} = $ts;

  if ($buffer eq "<jabber>  </jabber>") {
    #ping? pong!
    &jabber_output($FXsock{'jabber'}, " ");
    return(0);
  }

  my $parser = new XML::Parser('Style' => 'Tree');
  my $tree = $parser->parse($buffer);

  my $data = \@{$tree->[1]};
  shift(@{$data});

  if ($data->[0] eq "stream:error") {
    &putinfo("y", "Jabber stream error");
    &del_socket($FXsock{'jabber'});
    return(0);
  }

  if ($data->[0] eq "stream:stream") {
    my $username = (split(/\@/, $FXbot{'PageServ'}{'jabber_id'}, 0))[0];

    $FXim{'jabber'}{'id'} = ${$data->[1]}[0]->{'id'};

    &jabber_output($FXsock{'jabber'}, "<iq id='checkLogin' type='get'><query xmlns='jabber:iq:auth'><username>%s</username></query></iq>", $username);
    return(0);

  } elsif ($data->[0] eq "iq") {
    if (${$data->[1]}[0]->{'type'} eq "error") {
      my $code = ${$data->[1]}[4][0]->{'code'};

      &putinfo("y", "Jabber error: $jabber_error{$code} ($code)");

      &del_socket($FXsock{'jabber'});
      return(0);
    }

    if ((${$data->[1]}[0]->{'id'} eq "checkLogin") && (${$data->[1]}[0]->{'type'} eq "result")) {
      &jabber_id("digest");
      return(0);

    } elsif ((${$data->[1]}[0]->{'id'} eq "doConnect") && (${$data->[1]}[0]->{'type'} eq "result")) {
      &jabber_output($FXsock{'jabber'}, "<iq type='get'><query xmlns='jabber:iq:private'><roster xmlns='roster:delimiter'/></query></iq>");
      &jabber_output($FXsock{'jabber'}, "<iq type='get' id='buddies'><query xmlns='jabber:iq:roster'/></iq>");
      return(0);
    }

    my $presence = 0;
    for (my $idx = 1; $idx < scalar(@{$data}); $idx++) {
      if ($data->[$idx] eq "presence") {
        $presence = 1;
        next;
      }

      if ($presence) {
        my ($id, $resource) = split(/\//, $data->[$idx][0]->{'from'}, 2);

        if ($FXserv{'connected'} && $FXim{'jabber'}{'buddies'}{lc($id)} && $FXim{'jabber'}{'buddies'}{lc($id)}{'buddy'} && !$FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'}) { &jabber_introduce($id); }
        $presence = 0;
      }
    }

    if (${$data->[1]}[0]->{'id'} && (${$data->[1]}[0]->{'id'} eq "buddies")) {
      for (my $idx = 2; $idx < scalar(@{${$data->[1]}[2]}); $idx += 2) {
        my $jid = $data->[1][2][$idx][0]->{'jid'};
        my $sub = $data->[1][2][$idx][0]->{'subscription'};

        foreach (keys(%{$FXim{'jabber'}{'buddies'}})) {
          if ((lc($jid) eq lc($FXim{'jabber'}{'buddies'}{$_}{'id'})) && $FXim{'jabber'}{'buddies'}{$_}{'buddy'}) {
            if (($sub eq "none") || ($sub eq "from")) { &jabber_output($FXsock{'jabber'}, "<presence to='%s' type='subscribe'/>", $jid); }
            &jabber_output($FXsock{'jabber'}, "<presence to='%s'><priority>1</priority></presence>", $jid);
            last;
          }
        }
      }
    }

    if (${$data->[1]}[0]->{'type'} eq "result") { &jabber_output($FXsock{'jabber'}, "<presence><priority>1</priority></presence>"); }
    return(0);

  } elsif ($data->[0] eq "presence") {
    for (my $idx = 1; $idx < scalar(@{$data}); $idx += 2) {
      my ($id, $resource) = split(/\//, ${$data->[$idx]}[0]->{'from'}, 2);
      my $type = "both";

      if (${$data->[$idx]}[0]->{'type'}) { $type = ${$data->[$idx]}[0]->{'type'} };

      if (($type eq "unavailable") || ($type eq "unsubscribe") || ($type eq "unsubscribed")) {
        foreach (keys(%{$FXim{'jabber'}{'buddies'}})) {
          if ((lc($id) eq lc($FXim{'jabber'}{'buddies'}{$_}{'id'})) && $FXim{'jabber'}{'buddies'}{$_}{'buddy'}) {
            &jabber_disconnect($id, "Signed off");
            last;
          }
        }

      } elsif (($type eq "both") || ($type eq "subscribed")) {
        foreach (keys(%{$FXim{'jabber'}{'buddies'}})) {
          if (lc($id) eq lc($FXim{'jabber'}{'buddies'}{$_}{'id'})) {
            &jabber_output($FXsock{'jabber'}, "<presence to='%s'><priority>1</priority></presence>", $id);

            $FXim{'jabber'}{'buddies'}{$_}{'online'} = 1;
            $FXim{'jabber'}{'buddies'}{$_}{'info'} = $resource;
            $FXim{'jabber'}{'buddies'}{$_}{'status'} = ${$data->[1]}[2][2];
            $FXim{'jabber'}{'buddies'}{$_}{'seen'} = $ts;

            if ($FXserv{'connected'} && $FXim{'jabber'}{'buddies'}{$_}{'buddy'} && !$FXim{'jabber'}{'buddies'}{$_}{'ircnick'}) { &jabber_introduce($id); }
            last;
          }
        }

      } elsif ($type eq "subscribe") {
        &jabber_output($FXsock{'jabber'}, "<presence to='%s' type='subscribed'/>", $id);
      }
    }

    return(0);

  } elsif (($data->[0] eq "message") && ($data->[1][0]->{'type'} eq "chat")) {
    my ($id, $resource) = split(/\//, $data->[1][0]->{'from'}, 2);
    my $buffer = "";

    foreach my $session (keys(%{$FXim{'jabber'}{'board'}})) {
      if (lc($FXim{'jabber'}{'board'}{$session}{'buddy'}) eq lc($id)) {
        my $body = 0;

        for (my $idx = 1; $idx < scalar(@{$data->[1]}); $idx++) {
          if ($data->[1][$idx] eq "body") {
            $body = 1;
            next;
          }

          if ($body) {
            $buffer = $data->[1][$idx][2];
            utf8::decode($buffer);
            $body = 0;
            last;
          }
        }

        &putsock($FXsock{'irc'}, ":%s PRIVMSG %s :%s", $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'}, $FXim{'jabber'}{'board'}{$session}{'channel'}, $buffer);
        last;
      }
    }

    return(0);
  }

  return(0);
}

sub jabber_introduce {
  my $id = shift;

  my $nick = $FXim{'jabber'}{'buddies'}{lc($id)}{'nick'};
  my $ircnick = $FXbot{'PageServ'}{'jabber_prefix'}; $ircnick =~ s/\%/$nick/;

  if ($FXirc{'client'}{lc($ircnick)}) {
    for (my $idx = 1; 1; $idx++) {
      $ircnick = $ircnick . $idx;
      if (!$FXirc{'client'}{lc($ircnick)}) { last; }
    }
  }

  my $umode = ''; $umode |= $FXflags{'client'}{'i'};
  my $rand = "";

  $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'} = $ircnick;
  $FXim{'jabber'}{'buddies'}{lc($id)}{'online'} = 1;

  for (my $idx = 0; $idx < 6; $idx++) { $rand .= chr(97 + int(rand(26))); }

  my ($user, $domain) = split(/\@/, $id, 2);
  my $host = $rand . "." . $domain;
  my $info = $FXim{'jabber'}{'buddies'}{lc($id)}{'info'};

  &introduce_nick($ircnick, $nick, $host, undef, $umode, $info);
  &add_client($ircnick, $nick, $host, $info, undef, $FXserv{'name'}, $ts, $umode);

  &jabber_sjoin($id, $FXbot{'PageServ'}{'jabber_channel'}, 0);
  return(0);
}

sub jabber_disconnect {
  my ($id, $reason) = (shift, shift);

  my $ircnick = $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'};

  if ($ircnick && $FXserv{'connected'}) { 
    foreach my $session (keys(%{$FXim{'jabber'}{'board'}})) {
      if (lc($FXim{'jabber'}{'board'}{$session}{'buddy'}) eq lc($id)) {
        &jabber_remsession($session, "*");
        last;
      }
    }

    &add_seen($ircnick, "QUIT", $reason);
    &chanlog("q", undef, $ircnick, $reason, undef); 
    &putsock($FXsock{'irc'}, ":%s QUIT :%s", $ircnick, $reason);
    &del_client($ircnick);
  }

  $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'} = "";
  $FXim{'jabber'}{'buddies'}{lc($id)}{'seen'} = $ts;
  $FXim{'jabber'}{'buddies'}{lc($id)}{'status'} = "";
  $FXim{'jabber'}{'buddies'}{lc($id)}{'online'} = 0;

  return(0);
}

sub jabber_addbuddy {
  my ($id, $nick) = (shift, shift);
  my $done = 0;

  foreach (keys(%{$FXim{'jabber'}{'buddies'}})) {
    if ((lc($id) eq lc($FXim{'jabber'}{'buddies'}{$_}{'id'})) && !$FXim{'jabber'}{'buddies'}{$_}{'buddy'}) {
      $FXim{'jabber'}{'buddies'}{lc($id)}{'buddy'} = 1;
      $FXim{'jabber'}{'buddies'}{lc($id)}{'nick'} = $nick;

      if ($FXserv{'connected'} && $FXim{'jabber'}{'buddies'}{lc($id)}{'online'}) { &jabber_introduce($id); }

      $done = 1;
      last;
    }
  }

  if (!$done) {
    $FXim{'jabber'}{'buddies'}{lc($id)}{'id'} = $id;
    $FXim{'jabber'}{'buddies'}{lc($id)}{'nick'} = $nick;
    $FXim{'jabber'}{'buddies'}{lc($id)}{'info'} = "";
    $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'} = "";
    $FXim{'jabber'}{'buddies'}{lc($id)}{'online'} = 0;
    $FXim{'jabber'}{'buddies'}{lc($id)}{'status'} = "";
    $FXim{'jabber'}{'buddies'}{lc($id)}{'seen'} = 0;
    $FXim{'jabber'}{'buddies'}{lc($id)}{'buddy'} = 1;
  }

  &jabber_output($FXsock{'jabber'}, "<presence to='%s' type='subscribe'/>", $id);
  return(0);
}

sub jabber_rembuddy {
  my $id = shift;

  &jabber_output($FXsock{'jabber'}, "<presence to='%s' type='unsubscribe'/>", $id);

  &jabber_disconnect($id, "Removed");
  $FXim{'jabber'}{'buddies'}{lc($id)}{'buddy'} = 0;

  return(0);
}

sub jabber_privmsg {
  my ($nick, $id, $query) = (shift, shift, shift);

  if (&check_flood($nick)) { return(0); }

  if ($query =~ /\cA(.*)\cA/) {
    my $ctcp = uc((split(/ /, $1, 0))[0]);

    if ($ctcp eq "VERSION") {
      &buddy_ctcp($nick, $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'}, "VERSION %s", "FXserv IRC Services $FXserv{'version'}");
    } elsif ($ctcp eq "PING") {
      &buddy_ctcp($nick, $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'}, "PING %s", (split(/ /, $1, 0))[1]);
    } elsif ($ctcp eq "TIME") {
      &buddy_ctcp($nick, $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'}, "TIME %s", scalar(localtime()));
    }

    return(0);
  }

  my $done = 0;

  foreach my $session (keys(%{$FXim{'jabber'}{'board'}})) {
    if (lc($FXim{'jabber'}{'board'}{$session}{'buddy'}) eq lc($id)) {
      if (grep(lc($_) eq lc($nick), @{$FXim{'jabber'}{'board'}{$session}{'nick'}})) {
        &putsock($FXsock{'irc'}, ":%s %s %s :$FXmsg{'buddy_channel'}", $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'}, $FXcfg{'general'}{'communicate'}, $nick, $FXim{'jabber'}{'board'}{$session}{'channel'});
        $done = 1;
        last;
      } else {
        &putsock($FXsock{'irc'}, ":%s %s %s :$FXmsg{'buddy_request'}", $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'}, $FXcfg{'general'}{'communicate'}, $nick, $FXim{'jabber'}{'board'}{$session}{'channel'});
        &putsock($FXsock{'irc'}, ":%s INVITE %s %s", $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'}, $nick, $FXim{'jabber'}{'board'}{$session}{'channel'});
        $done = 1;
        last;
      }
    }
  }

  if (!$done) {
    &putsock($FXsock{'irc'}, ":%s %s %s :%s", $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'}, $FXcfg{'general'}{'communicate'}, $nick, $FXmsg{'buddy_invite'});

    my $session = $FXim{'jabber'}{'session'}++;

    $FXim{'jabber'}{'board'}{$session}{'session'} = $session;
    $FXim{'jabber'}{'board'}{$session}{'id'} = $id;
    $FXim{'jabber'}{'board'}{$session}{'buddy'} = $id;
    @{$FXim{'jabber'}{'board'}{$session}{'nick'}} = ();
    $FXim{'jabber'}{'board'}{$session}{'channel'} = "#jabber" . "." . $ts;
    $FXim{'jabber'}{'board'}{$session}{'start'} = $ts;
    $FXim{'jabber'}{'board'}{$session}{'last'} = $ts;

    &jabber_sjoin($id, $FXim{'jabber'}{'board'}{$session}{'channel'}, 1);

    &putsock($FXsock{'irc'}, ":%s %s %s :$FXmsg{'buddy_request'}", $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'}, $FXcfg{'general'}{'communicate'}, $nick, $FXim{'jabber'}{'board'}{$session}{'channel'});
    &putsock($FXsock{'irc'}, ":%s INVITE %s %s", $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'}, $nick, $FXim{'jabber'}{'board'}{$session}{'channel'});
    return(0);
  }

  return(0);
}

sub jabber_renamebuddy {
  my ($id, $nick) = (shift, shift);

  if ($FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'}) {
    my $newnick = $FXbot{'PageServ'}{'jabber_prefix'}; $newnick =~ s/\%/$nick/;
    my $oldnick = $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'};

    &putsock($FXsock{'irc'}, ":%s NICK %s", $oldnick, $newnick);
    &add_seen($oldnick, "NICK", $newnick);
    &chanlog("n", undef, $oldnick, $newnick, undef); 
    &rename_client($oldnick, $newnick);

    $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'} = $newnick;
  }

  $FXim{'jabber'}{'buddies'}{lc($id)}{'nick'} = $nick;

  return(0);
}

sub jabber_text {
  my ($id, $format, @arg) = (shift, shift, @_);
  my $buffer = sprintf($format, @arg);

  utf8::encode($buffer);

  $buffer =~ s/&/&amp;/g;
  $buffer =~ s/</&lt;/g;
  $buffer =~ s/>/&gt;/g;

  &jabber_output($FXsock{'jabber'}, "<message to='%s' type='chat'><body>%s</body></message>", $id, $buffer);
  return(0);
}

sub jabber_remsession {
  my ($session, $nick) = (shift, shift);
  my $id = $FXim{'jabber'}{'board'}{$session}{'id'};

  if ($nick && ($nick ne "*")) {
    my $cnt = 0;

    foreach my $user (@{$FXim{'jabber'}{'board'}{$session}{'nick'}}) {
      if (lc($nick) eq lc($user)) {
        splice(@{$FXim{'jabber'}{'board'}{$session}{'nick'}}, $cnt, 1);
        &jabber_text($FXim{'jabber'}{'board'}{$session}{'id'}, $FXmsg{'buddy_part'}, $nick);
        last;
      }

      $cnt++;
    }
  }

  if (($nick eq "*") || !scalar(@{$FXim{'jabber'}{'board'}{$session}{'nick'}})) {
    my $ircnick = $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'};
    my $channel = $FXim{'jabber'}{'board'}{$session}{'channel'};
    &putsock($FXsock{'irc'}, ":%s PART %s", $ircnick, $channel);

    &add_seen($ircnick, "PART", $channel);
    &chanlog("p", $channel, $ircnick, undef, undef);

    if (scalar(@{$FXim{'jabber'}{'board'}{$session}{'nick'}})) {
      foreach my $nick (@{$FXim{'jabber'}{'board'}{$session}{'nick'}}) { &jabber_text($id, $FXmsg{'buddy_part'}, $nick); }
    }

    delete($FXirc{'channel'}{lc($channel)}{'names'}{lc($ircnick)});
    &del_channel($channel);

    delete($FXim{'jabber'}{'board'}{$session});
  }

  return(0);
}

sub jabber_sjoin {
  my ($id, $channel, $session) = (shift, shift, shift);

  my $channel_ts = $ts;
  my $channel_mode = "";
  my $mode = '';
  my $prefix = "";

  my $nick = $FXim{'jabber'}{'buddies'}{lc($id)}{'nick'};
  my $ircnick = $FXim{'jabber'}{'buddies'}{lc($id)}{'ircnick'};

  if ($FXirc{'channel'}{lc($channel)}) {
    $channel_ts = $FXirc{'channel'}{lc($channel)}{'time'};
  } else {
    $mode |= $FXflags{'channel'}{'n'};
    $mode |= $FXflags{'channel'}{'t'};
    $channel_mode = "nt";

    if ($session) {
      $mode |= $FXflags{'channel'}{'i'};
      $mode |= $FXflags{'channel'}{'s'};
      $channel_mode .= "is";
    }

    $channel_ts = $ts;
    &add_channel($channel, $ts, $mode);
  }

  $FXirc{'channel'}{lc($channel)}{'names'}{lc($ircnick)}{'nick'} = $ircnick;
  $FXirc{'channel'}{lc($channel)}{'names'}{lc($ircnick)}{'mode'} = '';

  if ($session) {
    $prefix = "\@";
    $FXirc{'channel'}{lc($channel)}{'names'}{lc($ircnick)}{'mode'} |= $FXflags{'channel'}{'o'};
  }

  if (++$FXirc{'channel'}{lc($channel)}{'current'} > $FXirc{'channel'}{lc($channel)}{'max'}) {
    $FXirc{'channel'}{lc($channel)}{'max'} = $FXirc{'channel'}{lc($channel)}{'current'};
    $FXirc{'channel'}{lc($channel)}{'max_ts'} = $ts;
  }

  &add_seen($ircnick, "JOIN", $channel);
  &chanlog("j", $channel, $ircnick, undef, undef); 
  &putsock($FXsock{'irc'}, ":%s SJOIN %d %s +%s :%s%s", $FXserv{'name'}, $channel_ts, $channel, $channel_mode, $prefix, $ircnick);

  if ($session) {
    &chanlog("m", $channel, $FXserv{'name'}, "+o", $ircnick);

    if (!($FXirc{'channel'}{lc($channel)}{'mode'} & $FXflags{'channel'}{'i'})) {
      &channel_mode($ircnick, $channel, "+i", undef);
      &chanlog("m", $channel, $ircnick, "+i", undef);
    }

    if (!($FXirc{'channel'}{lc($channel)}{'mode'} & $FXflags{'channel'}{'s'})) {
      &channel_mode($ircnick, $channel, "+s", undef);
      &chanlog("m", $channel, $ircnick, "+s", undef);
    }
  }

  return(0);
}

1;